#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <omp.h>

int BLOQUE,THREADS;
/*
c
c Multiplicacion matriz-vector, no optimizada, para probar OpenMP
22-11-2004
c
*/

void mv(double *a, int fa,int ca,int lda,double *b,int fb,double *c,int fc)
{
  int i, j;
  double s;
  #pragma omp parallel 
  {
    #pragma omp master
    {
      THREADS=omp_get_num_threads();
    }
#ifdef DEBUG
#pragma omp critical
{
    printf("Thread %d\n",omp_get_thread_num());
}
#endif
    #pragma omp for private(i,j,s) schedule(static,BLOQUE)
    for (i = 0; i < fa; i++) 
    {
      #ifdef DEBUG
        #pragma omp critical
        {
          printf("Fila %d, thread %d\n",i,omp_get_thread_num());
        }
      #endif
      s=0.;
      for(j=0;j<ca;j++)
      {
        s+=a[i*lda+j]*b[j];
      }
      c[i]=s;
    }
  }
}

/*Multiplicacion matriz-matriz*/

void mm(double *a, int fa,int ca,int lda,double *b,int fb,int cb,int ldb,double *c,int fc,int cc,int ldc)
{
  int i, j, k;
  double s;

  #pragma omp parallel 
  {
#ifdef DEBUG
    #pragma omp critical
    {
        printf("Thread %d\n",omp_get_thread_num());
    }
#endif
    #pragma omp for private(j,k,s) schedule(static,BLOQUE)
    for (i = 0; i < fa; i++) 
    {
      #ifdef DEBUG
        #pragma omp critical
        {
          printf("Fila %d, proceso %d\n",i,omp_get_thread_num());
        }
      #endif
      for(j=0;j<cb;j++)
      {
        s=0.;
        for (k = 0; k < ca; k++)
          s = s+a[i*lda+k]*b[k*ldb+j];
        c[i*ldc+j]=s;
      }
    }
  }
}
/*
c
c initialize - random initialization for array
c
*/

void initialize(double *m, int f,int c,int ld){
  int i, j;

  for (i = 0; i < f; i++)
  {
    for (j = 0; j < c; j++)
    {  
      m[i*ld+j] = (double)(i+j);
    }
  }
}

void initializealea(double *m, int f,int c,int ld){
  int i, j;

//#pragma omp parallel for schedule(static) private(i,j) 
  for (i = 0; i < f; i++)
  {
    for (j = 0; j < c; j++)
    {  
      m[i*ld+j] = (double)rand()/RAND_MAX;
    }
  }
}

void escribir(double *m, int f,int c,int ld){
  int i, j;

  for (i = 0; i < f; i++)
  {
    for (j = 0; j < c; j++)
    {  
      printf("%.4lf ",m[i*ld+j]);
    }
    printf("\n");
  }
}
/*
c
c     mseconds - returns elapsed microseconds since Jan 1st, 1970.
c
*/
double mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000000. + t.tv_usec;
}

int main(int argc,char *argv[]) {
  int N,i, j,fa,ca,lda,fb,cb,ldb,fc,cc,ldc,nm,iN,fN,incN,iB,fB,incB;
  double start,fin,tiempo,Mflops;
  double *a,*b,*c;

  iN=atoi(argv[1]);
  fN=atoi(argv[2]);
  incN=atoi(argv[3]);
  iB=atoi(argv[4]);
  fB=atoi(argv[5]);
  incB=atoi(argv[6]);

  for(N=iN;N<fN;N+=incN)
  {
    for(BLOQUE=iB;BLOQUE<fB;BLOQUE+=incB)
    {
      fa=ca=lda=fb=cb=ldb=fc=cc=ldc=N;

      a = (double *) malloc(sizeof(double)*fa*ca);
      b = (double *) malloc(sizeof(double)*fb*cb);
      c = (double *) malloc(sizeof(double)*fc*cc);
   
      initializealea(a,fa,ca,lda);
      initializealea(b,fb,cb,ldb);
 
      start = mseconds();

      nm=10;
      for(i=0;i<nm;i++)
        mv(a,fa,ca,lda,b,fb,c,fc);

      fin=mseconds();
      tiempo=((fin-start)/nm)/1000000.;
      if(tiempo==0.)
      {
        printf("No hay suficiente precision\n");
      }
      else
      {
        Mflops=((2.*N*N)/tiempo)/1000000.;
        printf("Multiplicaciones MATRIZ-VECTOR\n  bloque %d, threads %d\n    en seconds: %.6lf, Mflops: %.6lf, Mflops por thread: %.6lf\n",BLOQUE,THREADS,tiempo,Mflops,Mflops/THREADS);
        printf("c[%d]=%.8lf\n",fc,c[fc-1]);
      }

      start = mseconds();

      nm=1;
      for(i=0;i<nm;i++)
        mm(a,fa,ca,lda,b,fb,cb,ldb,c,fc,cc,ldc);

      fin=mseconds();
      tiempo=((fin-start)/nm)/1000000.;
      if(tiempo==0.)
      {
        printf("No hay suficiente precision\n");
      }
      else
      {
        Mflops=((2.*N*N*N)/tiempo)/1000000.;
        printf("Multiplicacion MATRIZ-MATRIZ\n  bloque %d, threads %d\n    en seconds: %.6lf, Mflops: %.6lf, Mflops por thread: %.6lf\n",BLOQUE,THREADS,tiempo,Mflops,Mflops/THREADS);
        printf("c[%d,%d]=%.8lf\n",fc,cc,c[(fc-1)*ldc+cc-1]);
      }

      free(a);
      free(b);
      free(c);
    }
  }
}
